"""
SdeClient
Created: 2020-04-22
Last Updated: 2023-03-07
Author: Scott McDowell (smcdowell@securitycompass.com)

#region Change Log

2020-04-22 - Initial version.
2020-05-15 - Added get_dictionary_ methods.
2020-06-05 - Added get_api_request_count and raise_for_status() calls to patch
             and post class methods.
2020-06-18 - Moved json() method calls to return statements (after
             raise_for_status) in patch and post class methods.
2020-07-10 - Moved get_dictionary_from_results method to SdeUtilities.
2020-12-04 - Added get_projects, add_business_unit_to_projects, and 
             get_project_tasks helper methods.
2021-01-25 - Added URLs related to third-party libraries.
2021-02-25 - Added get_user_id_from_email.
             Moved page count calculation used for pagination to its own method.
             Added support for requests library retries to make lengthy
             paginated requests more likely to succeed. (Also switched to using
             HTTPAdapter class methods as a result.)
2021-05-05 - Added get_project_survey_data.
2021-06-02 - Added more change log entries to text header.
             Fall back to single-threaded pagination if ThreadPool import fails.
2021-06-23 - Added get_email_from_user_id.
2021-07-29 - Added create_project_release.
             Added options to get_respone_json, get_results, patch, and post
             methods to not call raise_for_status() method. This allows class
             users to parse an API error rather that getting a generic code
             error code. (Also added optional parameter to helper methods that
             use any of the four listed methods.)
2021-12-07 - Added optional proxy support. (Also see sdeclient_settings.yaml)
             Added default timeout.
2021-12-10 - Added delete method.
             Reorganized init method for readability.
             Set headers and verify in init method for entire session.
2022-01-19 - Added select_draft_answer method.
2022-02-04 - Added methods to lock/unlock a project and a project survey.
             Added get_project method.
2022-02-08 - Added clear_survey_draft and save_survey_draft methods.
2022-03-09 - Added synchronous export_library method which uses polling to wait
             for the export job to finish.
2022-04-05 - Replaced deprecated Retry parameter 'method_whitelist' with new
             'allowed_methods' parameter.
2022-04-06 - Added get_slim_projects method which obtains minimal project data
             from the applications endpoint and tends to be faster than
             using the projects endpoint if you only need a list of project
             names, IDs, and their relationship to BUs and applications.
2022-04-29 - Added archive_project and unarchive_project methods.
2022-05-27 - Added {"include": "survey,hidden"} to select_draft_answer,
             clear_survey_draft, and save_survey_draft methods to ensure that
             the entire survey contents are available in the response.
2022-06-29 - Modified get_project_survey_data to better handle hidden answers
             and expand answers to include properties in addition to the ID.
             Added get_library_survey, get_library_questions, and
             get_library_answers methods.
             Added get_project_activities method.
2022-08-30 - Added 'last_response' class property to hold the most recent API
             response so details such as API errors can be obtained.
2022-12-13 - Added get_application, get_applications, archive_application, and
             unarchive_application methods.
             Modified get_project and get_projects to include optional
             'archived' parameter to align with new application methods.
2022-12-16 - Added support for querying the SDE Cube API using the
             "get_cube_query_data" and "get_cube_report_data" methods.
2023-02-08 - Added get_project_task, add_project_task_from_library, 
             update_project_task, and get_library_tasks methods.
2023-02-27 - Removed 'include_archived' parameter from get_slim_projects
             because the applications 'include=projects' request does not
             include archived projects in its response.
2023-03-07 - Added get_business_units, get_business_unit, delete_project,
             delete_application, and delete_business_unit methods.

#endregion Change Log
"""

# Fall back to single-threaded pagination.
try:
    from multiprocessing.pool import ThreadPool
except:
    single_threaded_fallback = True
else:
    single_threaded_fallback = False

# https://pyyaml.org/
# https://pypi.org/project/PyYAML/
import yaml
# https://requests.readthedocs.io/
# https://pypi.org/project/requests/
import requests
from requests.adapters import HTTPAdapter
# Verification warnings are disabled for individual requests to avoid flooding
# the console. Add your own warning when you create an instance of SdeClient
# with verify=false to notify the user.
from requests.packages import urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
from urllib3.util.retry import Retry
# Used to shim requests module's Session 'request' function to set a default
# timeout value for all request methods.
# Reference: https://stackoverflow.com/a/59317604
import functools
# Sleep method used to implement polling for job completion.
import time
# Datetime used to determine whether the Cube API JWT requires a refresh.
from datetime import datetime
import json


from sdeutilities import sdeutils


class sde:
    """
    Simple SD Elements API client that uses the requests library.
    """

    SETTINGS_FILENAME = "sdeclient_settings.yaml"

    def __init__(self, server_url, api_token, verify):
        assert isinstance(server_url, str)
        assert isinstance(api_token, str)
        assert isinstance(verify, bool)

        # Read settings file.
        self.settings = yaml.safe_load(open(self.SETTINGS_FILENAME))

        # Use a single requests session for all API communication.
        # This allows for one-time configuration of some settings.
        self.session = requests.Session()

        # SDE server URL.
        self.server_url = (
            server_url.strip("/")
            if server_url.startswith("https://")
            else "https://" + server_url.strip("/")
        )

        # Request headers.
        self.session.headers = {
            "Authorization": f"Token {api_token}",
            "Content-Type": "application/json",
            "Accept": "application/json"
        }

        # Verify server certificate.
        self.verify_certificate = verify
        self.session.verify = self.verify_certificate

        # Proxy.
        self.proxies = {
            "http": self.settings["proxy"]["http"],
            "https": self.settings["proxy"]["https"]
        }
        self.session.proxies.update(self.proxies)

        # Paginated request page size.
        self.page_size = int(self.settings["pagination"]["page_size"])
        assert 10 <= self.page_size <= 1000

        # Paginated request threading.
        # Allow user to choose single_threaded by changing settings file to
        # specify one thread. May also need to apply single_threaded based on
        # success/failure of multiprocessing.pool import statement.
        self.thread_count = int(self.settings["pagination"]["thread_count"])
        assert 1 <= self.thread_count <= 16
        if self.thread_count == 1 or single_threaded_fallback:
            self.single_threaded = True
            self.thread_count = 1
        else:
            self.single_threaded = False

        # Timeout.
        # Note that timeouts can cause the client to retry so it may take
        # multiple timeout periods for a timeout error to appear in terminal.
        timeout_connect = int(self.settings["timeout"]["connect"])
        assert 5 <= timeout_connect <= 600
        timeout_read = int(self.settings["timeout"]["read"])
        assert 5 <= timeout_read <= 600
        self.session.request = functools.partial(
            self.session.request, timeout=(timeout_connect, timeout_read))

        # Retries.
        retry_count = self.settings["retry"]["total"]
        assert 1 <= retry_count <= 10
        self.retry_strategy = Retry(
            total=retry_count,
            backoff_factor=self.settings["retry"]["backoff_factor"],
            status_forcelist=self.settings["retry"]["status_forcelist"],
            allowed_methods=self.settings["retry"]["allowed_methods"]
        )
        adapter = HTTPAdapter(max_retries=self.retry_strategy)
        self.session.mount("https://", adapter)

        # Library export.
        self.library_export_delay = int(
            self.settings["library"]["export_delay"])
        assert 5 <= self.library_export_delay <= 60
        self.library_export_retries = int(
            self.settings["library"]["export_retries"])
        assert 2 <= self.library_export_retries <= 60
        assert (self.library_export_delay * self.library_export_retries) <= 600

        # API endpoints.
        self.endpoints = self.settings["endpoints"]

        # Count API interactions for optimization testing.
        self.api_request_count = 0

        # Used to store user emails to prevent repeated API calls for the same
        # user ID.
        self.user_email_cache = {}

        # Global role cache used for optimization when making repeated requests.
        self.global_role_cache = {}

        # Retain most recent response object to provide access to error info.
        self.last_response = None

        # Retain Cube API JWT information.
        ## self.cube_session = None ?? refactor SdeClient to have two sessions ??
        self.cube_jwt = None
        self.cube_auth_endpoint = self.settings["cube_api"]["auth_endpoint"]
        self.cube_query_endpoint = self.settings["cube_api"]["query_endpoint"]
        self.cube_jwt_max_age = self.settings["cube_api"]["jwt_max_age"]
        self.cube_jwt_updated = None

    def get_response_json(self, endpoint, parameters, raise_for_status=True):
        """
        GET and return response JSON.
        """
        response = self.session.get(
            self.server_url + endpoint,
            params=parameters
        )
        self.api_request_count += 1
        self.last_response = response
        if raise_for_status:
            response.raise_for_status()
        return response.json()

    def patch(self, endpoint, parameters, json_payload, raise_for_status=True):
        """
        PATCH and return response JSON.
        """
        response = self.session.patch(
            self.server_url + endpoint,
            params=parameters,
            json=json_payload
        )
        self.api_request_count += 1
        self.last_response = response
        if raise_for_status:
            response.raise_for_status()
        return response.json()

    def post(self, endpoint, parameters, json_payload, raise_for_status=True):
        """
        POST and return response JSON.
        """
        response = self.session.post(
            self.server_url + endpoint,
            params=parameters,
            json=json_payload
        )
        self.api_request_count += 1
        self.last_response = response
        if raise_for_status:
            response.raise_for_status()
        return response.json()

    def delete(self, endpoint, parameters, raise_for_status=True):
        """
        DELETE. Nothing to return.
        """
        response = self.session.delete(
            self.server_url + endpoint,
            params=parameters
        )
        self.api_request_count += 1
        self.last_response = response
        if raise_for_status:
            response.raise_for_status()

    def get_item_count(self, endpoint, parameters):
        """
        Get count portion of response by supplying a page size of 1.
        """
        local_parameters = parameters.copy()
        local_parameters["page_size"] = ["1"]
        local_parameters["page"] = ["1"]
        response_json = self.get_response_json(endpoint, local_parameters)
        return response_json["count"]

    def get_page_count(self, item_count):
        """
        Get page count required for provided item count.
        """
        page_count = (
            (item_count // self.page_size)
            + (1 if item_count % self.page_size > 0 else 0)
        )
        return page_count

    def get_results(self, endpoint, parameters, raise_for_status=True):
        """
        Get results portion of response.
        """
        response_json = self.get_response_json(
            endpoint, parameters, raise_for_status=raise_for_status)
        return response_json["results"]

    def get_paginated_results(self, endpoint, parameters):
        """
        Make paginated requests and return combined results.
        """
        # Count data pages.
        local_parameters = parameters.copy()
        item_count = self.get_item_count(endpoint, local_parameters)
        page_count = self.get_page_count(item_count)
        local_parameters["page_size"] = [str(self.page_size)]

        if self.single_threaded:
            return self._get_paginated_results_single_threaded(
                endpoint, local_parameters, page_count)
        else:
            return self._get_paginated_results_multi_threaded(
                endpoint, local_parameters, page_count)

    def _get_paginated_results_multi_threaded(
        self, endpoint, parameters, page_count):
        """
        Make paginated asynchronous requests and return combined results.
        """

        # Use multiple threads to call get_results and append response data to
        # results list.
        async_results = []
        request_pool = ThreadPool(self.thread_count)
        for page_number in range(1, page_count + 1):
            thread_parameters = parameters.copy()
            thread_parameters["page"] = [str(page_number)]
            async_results.append(
                request_pool.apply_async(
                    self.get_results,
                    args=(endpoint, thread_parameters)
                )
            )
        request_pool.close()
        request_pool.join()

        # Return list of objects from the complete list of results.
        return [item for result in async_results for item in result.get()]

    def _get_paginated_results_single_threaded(
        self, endpoint, parameters, page_count):
        """
        Make paginated serialized requests and return combined results.
        """

        # Use single thread to call get_results and append response data to 
        # results list.
        results = []
        for page_number in range(1, page_count + 1):
            request_parameters = parameters.copy()
            request_parameters["page"] = [str(page_number)]
            response = self.get_results(endpoint, request_parameters)
            results += response
        
        return results

    def get_results_as_dictionary(self, endpoint, parameters):
        """
        Get results portion of response and return dictionary with "id" value
        as the key.
        """
        results = self.get_results(endpoint, parameters)
        return sdeutils.get_dictionary_from_collection_with_ids(results)

    def get_paginated_results_as_dictionary(self, endpoint, parameters):
        """
        Make paginated asynchronus requests and return combined results in a
        dictionary with "id" value as the key.
        """
        results = self.get_paginated_results(endpoint, parameters)
        return sdeutils.get_dictionary_from_collection_with_ids(results)

    def get_api_request_count(self):
        """
        Get number of API requests made by instace.
        """
        return self.api_request_count


# Helper Methods
# Commonly used requests added here for one-line convenience. These methods
# rely on the core class functionality above.

    def get_project(self, id, expand="", include="", exclude="", query={}, archived=False):
        """
        Get a single project. Optional expand, include, and exclude strings are
        used for URL parameter values. Optional query dict used to apply
        individual query parameters.
        """
        parameters = {}
        if expand:
            parameters["expand"] = expand
        if include:
            parameters["include"] = include
        if exclude:
            parameters["exclude"] = exclude
        if query:
            for key, value in query.items():
                parameters[key] = str(value)

        if archived:
            endpoint = (self.endpoints["archived_project"]
                            .replace("{archived_project_id}", str(id)))
        else:
            endpoint = (self.endpoints["project"]
                            .replace("{project_id}", str(id)))

        return self.get_response_json(endpoint=endpoint, parameters=parameters)

    def delete_project(self, id):
        """
        Delete a project.
        """
        return self.delete(
            endpoint=self.endpoints["project"].replace("{project_id}", str(id)),
            parameters={}
        )

    def get_application(self, id, expand="", include="", exclude="", query={}, archived=False):
        """
        Get a single application. Optional expand, include, and exclude strings
        are used for URL parameter values. Optional query dict used to apply
        individual query parameters.
        """
        parameters = {}
        if expand:
            parameters["expand"] = expand
        if include:
            parameters["include"] = include
        if exclude:
            parameters["exclude"] = exclude
        if query:
            for key, value in query.items():
                parameters[key] = str(value)

        if archived:
            endpoint = (self.endpoints["archived_application"]
                            .replace("{archived_application_id}", str(id)))
        else:
            endpoint = (self.endpoints["application"]
                            .replace("{application_id}", str(id)))

        return self.get_response_json(endpoint=endpoint, parameters=parameters)

    def delete_application(self, id):
        """
        Delete an application.
        """
        return self.delete(
            endpoint=self.endpoints["application"]
                         .replace("{application_id}", str(id)),
            parameters={}
        )

    def get_business_unit(self, id, include=""):
        """
        Get a single business unit. Optional include string.
        """
        parameters = {}
        if include:
            parameters["include"] = include

        return self.get_response_json(
            endpoint=self.endpoints["business_unit"]
                         .replace("{business_unit_id}", str(id)),
            parameters=parameters
        )
    
    def delete_business_unit(self, id):
        """
        Delete a business unit.
        """
        return self.delete(
            endpoint=self.endpoints["business_unit"]
                         .replace("{business_unit_id}", str(id)),
            parameters={}
        )

    def get_project_activities(self, id, query={}):
        """
        Get project activities.
        """
        parameters = {}
        if query:
            for key, value in query.items():
                parameters[key] = str(value)
        return self.get_paginated_results(
            self.endpoints["project_activities"]
                .replace("{project_id}", str(id)),
            parameters=parameters
        )

    def archive_project(self, project_id):
        """
        Archive a project by setting its archived attribute to True.
        """
        response_json = self.patch(
            self.endpoints["project"].replace("{project_id}", str(project_id)),
            parameters={},
            json_payload={"archived": True}
        )
        return response_json

    def archive_application(self, application_id, parameters={}):
        """
        Archive an application by setting its archived attribute to True.
        """
        response_json = self.patch(
            endpoint=self.endpoints["application"]
                         .replace("{application_id}", str(application_id)),
            parameters=parameters,
            json_payload={"archived": True}
        )
        return response_json

    def unarchive_project(self, project_id):
        """
        Unarchive a project by setting its archived attribute to False.
        """
        response_json = self.patch(
            self.endpoints["archived_project"]
                .replace("{archived_project_id}", str(project_id)),
            parameters={},
            json_payload={"archived": False}
        )
        return response_json

    def unarchive_application(self, application_id, parameters={}):
        """
        Unarchive an application by setting its archived attribute to False.
        """
        response_json = self.patch(
            endpoint=(
                self.endpoints["archived_application"]
                    .replace("{archived_application_id}", str(application_id))
            ),
            parameters=parameters,
            json_payload={"archived": False}
        )
        return response_json

    def get_projects(self, expand="", include="", exclude="", query={}, archived=False):
        """
        Get projects. Optional expand, include, and exclude strings are used
        for URL parameter values. Optional query dict used to apply individual
        query parameters.
        """
        parameters = {}
        if expand:
            parameters["expand"] = expand
        if include:
            parameters["include"] = include
        if exclude:
            parameters["exclude"] = exclude
        if query:
            for key, value in query.items():
                parameters[key] = str(value)

        if archived:
            endpoint = self.endpoints["archived_projects"]
        else:
            endpoint = self.endpoints["projects"]

        return self.get_paginated_results(
            endpoint=endpoint,
            parameters=parameters
        )

    def get_applications(self, expand="", include="", exclude="", query={}, archived=False):
        """
        Get applications. Optional expand, include, and exclude strings are used
        for URL parameter values. Optional query dict used to apply individual
        query parameters.
        """
        parameters = {}
        if expand:
            parameters["expand"] = expand
        if include:
            parameters["include"] = include
        if exclude:
            parameters["exclude"] = exclude
        if query:
            for key, value in query.items():
                parameters[key] = str(value)

        if archived:
            endpoint = self.endpoints["archived_applications"]
        else:
            endpoint = self.endpoints["applications"]

        return self.get_paginated_results(
            endpoint=endpoint,
            parameters=parameters
        )

    def get_slim_projects(self):
        """
        Get minimal project data from the applications endpoint. Includes BU,
        application, and project name/ID. This tends to be faster than using
        the projects endpoint if obtaining project IDs for use in other API
        requests.
        """
        applications = self.get_paginated_results(
            self.endpoints["applications"], {"include": "projects"})
        slim_projects = [
            {
                "id": project["id"],
                "name": project["name"],
                "application_id": application["id"],
                "application_name": application["name"],
                "business_unit_id": application["business_unit"]["id"],
                "business_unit_name": application["business_unit"]["name"]
            }
            for application in applications
            for project in application["projects"]
        ]
        return slim_projects

    def get_business_units(self, include="", query={}):
        """
        Get business units. Optional include string is used for URL parameter
        values. Optional query dict used to apply individual query parameters.
        """

        parameters = {}
        if include:
            parameters["include"] = include
        if query:
            for key, value in query.items():
                parameters[key] = str(value)

        return self.get_paginated_results(
            endpoint=self.endpoints["business_units"],
            parameters=parameters
        )

    def add_business_unit_to_projects(self, projects):
        """
        Adds 'business_unit' dictionary containing the project's business unit
        data from its parent application.
        """
        
        # Get applications by ID for easy lookup. Expand business unit.
        applications_by_id = (
            self.get_paginated_results_as_dictionary(
                self.endpoints["applications"],
                {"expand": "business_unit"}
            )
        )

        # Determine how to get application ID from projects.
        application_expanded = False
        if isinstance(projects[0]["application"], dict):
            application_expanded = True
        
        # Add business unit data.
        for project in projects:
            application_id = (project["application"]["id"]
                              if application_expanded
                              else project["application"])
            project["business_unit"] = (
                applications_by_id[application_id]["business_unit"]
            )

    def get_project_tasks(self, id, expand="", include="", filter={}):
        """
        Get tasks for supplied project ID. Optional expand and include strings
        are used for URL parameter values. Optional filter dict used to apply
        individual filter parameters. Returns all tasks regardless of accepted
        status if accepted/accepted__in filter parameter is not included.
        """
        parameters = {}
        if expand:
            parameters["expand"] = expand
        if include:
            parameters["include"] = include
        if filter:
            for key, value in filter.items():
                parameters[key] = str(value)

        # Apply accepted__in=true,false if no variation of the accepted filter
        # parameter has been included. This sidesteps the API's default
        # behaviour that only returns accepted tasks.
        if "accepted" not in parameters and "accepted__in" not in parameters:
            parameters["accepted__in"] = "true,false"

        return self.get_paginated_results(
            self.endpoints["project_tasks"].replace("{project_id}", str(id)),
            parameters
        )

    def get_project_task(
        self, project_id, task_id, expand="", include="", filter={}):
        """
        Get a single project task for supplied project ID and task ID. Task ID
        must be in the form of the library item ID (e.g. T123 or CT456).
        """
        parameters = {}
        if expand:
            parameters["expand"] = expand
        if include:
            parameters["include"] = include
        if filter:
            for key, value in filter.items():
                parameters[key] = str(value)

        project_task = self.get_response_json(
            self.endpoints["project_task"]
                .replace("{project_id}", str(project_id))
                .replace("{project_task_id}", f"{project_id}-{task_id}"),
            parameters=parameters
        )
        return project_task      

    def update_project_task(self, project_id, task_id, payload):
        """
        Patch a project task.
        """
        project_task = self.patch(
            endpoint=self.endpoints["project_task"]
                         .replace("{project_id}", str(project_id))
                         .replace("{project_task_id}", f"{project_id}-{task_id}"),
            parameters={},
            json_payload=payload
        )
        return project_task

    def get_user_id_from_email(self, email, raise_for_status=True):
        """
        Get user ID (int) by looking up supplied email and returning first
        matching result. Returns None if user not found.
        """
        user_id = None
        try:
            user = self.get_results(
                self.endpoints["users"],
                {"email": email},
                raise_for_status=raise_for_status
            )[0]
        except:
            pass
        else:
            if user:
                user_id = user["id"]
        return user_id

    def get_email_from_user_id(self, id, raise_for_status=True):
        """
        Get email address by looking up supplied user ID. Cache results to
        prevent repeated API requests for the same ID which could occur when
        adding email addresses to activity objects.
        """
        if id in self.user_email_cache:
            user_email = self.user_email_cache[id]
        else:
            user_email = self.get_response_json(
                self.endpoints["user"].replace("{user_id}", str(id)),
                {},
                raise_for_status=raise_for_status
            )["email"]
            self.user_email_cache[id] = user_email
        return user_email

    def get_library_tasks(
        self, priority=None, absolute_urls=None, show_inactive=None,
        expand="", include=""
    ):
        """
        Get library tasks as dictionary keyed by task item ID.
        """
        parameters = {}
        if priority:
            parameters["priority"] = priority
        if absolute_urls:
            parameters["absolute_urls"] = absolute_urls
        if show_inactive:
            parameters["show_inactive"] = show_inactive
        if expand:
            parameters["expand"] = expand
        if include:
            parameters["include"] = include
        library_tasks = self.get_paginated_results_as_dictionary(
            endpoint=self.endpoints["library_tasks"],
            parameters=parameters
        )
        return library_tasks

    def get_library_survey(self):
        """
        Get library survey by requesting the survey draft from an active project
        with the include=survey parameter.
        """

        # Get first active project returned by the API.
        project = self.get_results(
            endpoint=self.endpoints["projects"],
            parameters={"page_size": 1}
        )[0]

        # Request survey draft for the project and include the survey.
        project_survey_draft = self.get_response_json(
            endpoint=self.endpoints["project_survey_draft"]
                         .replace("{project_id}", str(project["id"])),
            parameters={"include": "survey,hidden"}
        )

        # Get library survey dictionary excluding project-specific fields.
        survey = {"sections": {}}
        for section in project_survey_draft["sections"]:
            (survey["sections"][section["id"]]) = (
                {
                    "title": section["title"],
                    "hidden": section["hidden"],
                    "subsections": {}
                }
            )
            for subsection in section["subsections"]:
                (survey["sections"][section["id"]]
                       ["subsections"][subsection["id"]]) = (
                    {
                        "title": subsection["title"],
                        "hidden": subsection["hidden"],
                        "questions": {}
                    }
                )
                for question in subsection["questions"]:
                    (survey["sections"][section["id"]]
                           ["subsections"][subsection["id"]]
                           ["questions"][question["id"]]) = (
                        {
                            "text": question["text"],
                            "description": question["description"],
                            "format": question["format"],
                            "mandatory": question["mandatory"],
                            "hidden": question["hidden"],
                            "answers": {},
                            "subquestions": {}
                        }
                    )
                    for answer in question["answers"]:
                        (survey["sections"][section["id"]]
                               ["subsections"][subsection["id"]]
                               ["questions"][question["id"]]
                               ["answers"][answer["id"]]) = (
                            {
                                "text": answer["text"],
                                "description": answer["description"],
                                "hidden": answer["hidden"]
                            }
                        )
                    for subquestion in question["subquestions"]:
                        (survey["sections"][section["id"]]
                               ["subsections"][subsection["id"]]
                               ["questions"][question["id"]]
                               ["subquestions"][subquestion["id"]]) = (
                            {
                                "text": subquestion["text"],
                                "description": subquestion["description"],
                                "format": subquestion["format"],
                                "mandatory": subquestion["mandatory"],
                                "hidden": subquestion["hidden"],
                                "answers": {}
                            }
                        )
                        for answer in subquestion["answers"]:
                            (survey["sections"][section["id"]]
                                   ["subsections"][subsection["id"]]
                                   ["questions"][question["id"]]
                                   ["subquestions"][subquestion["id"]]
                                   ["answers"][answer["id"]]) = (
                                {
                                    "text": answer["text"],
                                    "description": answer["description"],
                                    "hidden": answer["hidden"]
                                }
                            )

        return survey

    def get_library_questions(self):
        """
        Get library survey questions.
        """
        questions = {}
        survey = self.get_library_survey()
        for s_id, section in survey["sections"].items():
            for ss_id, subsection in section["subsections"].items():
                for q_id, question in subsection["questions"].items():
                    questions[q_id] = {
                        "text": question["text"],
                        "description": question["description"],
                        "format": question["format"],
                        "mandatory": question["mandatory"],
                        "hidden": question["hidden"],
                        "answers": question["answers"],
                        "subquestions": question["subquestions"],
                        "section": {
                            "id": s_id,
                            "title": section["title"],
                            "hidden": section["hidden"]
                        },
                        "subsection": {
                            "id": ss_id,
                            "title": subsection["title"],
                            "hidden": subsection["hidden"]
                        },
                        "parent_question": None
                    }
                    for sq_id, subquestion in question["subquestions"].items():
                        questions[sq_id] = {
                            "text": subquestion["text"],
                            "description": subquestion["description"],
                            "format": subquestion["format"],
                            "mandatory": subquestion["mandatory"],
                            "hidden": subquestion["hidden"],
                            "answers": subquestion["answers"],
                            "section": {
                                "id": s_id,
                                "title": section["title"],
                                "hidden": section["hidden"]
                            },
                            "subsection": {
                                "id": ss_id,
                                "title": subsection["title"],
                                "hidden": subsection["hidden"]
                            },
                            "parent_question": {
                                "id": q_id,
                                "text": question["text"],
                                "hidden": question["hidden"],
                            }
                        }
        return questions

    def get_library_answers(self):
        """
        Get library survey answers.
        """
        answers = {}
        survey = self.get_library_survey()
        for s_id, section in survey["sections"].items():
            for ss_id, subsection in section["subsections"].items():
                for q_id, question in subsection["questions"].items():
                    for a_id, answer in question["answers"].items():
                        answers[a_id] = {
                            "text": answer["text"],
                            "description": answer["description"],
                            "hidden": answer["hidden"],
                            "section": {
                                "id": s_id,
                                "title": section["title"],
                                "hidden": section["hidden"]
                            },
                            "subsection": {
                                "id": ss_id,
                                "title": subsection["title"],
                                "hidden": subsection["hidden"]
                            },
                            "parent_question": None,
                            "question": {
                                "id": q_id,
                                "text": question["text"],
                                "description": question["description"],
                                "format": question["format"],
                                "mandatory": question["mandatory"],
                                "hidden": question["hidden"]
                            }
                        }
                    for sq_id, subquestion in question["subquestions"].items():
                        for a_id, answer in subquestion["answers"].items():
                            answers[a_id] = {
                                "text": answer["text"],
                                "description": answer["description"],
                                "hidden": answer["hidden"],
                                "section": {
                                    "id": s_id,
                                    "title": section["title"],
                                    "hidden": section["hidden"]
                                },
                                "subsection": {
                                    "id": ss_id,
                                    "title": subsection["title"],
                                    "hidden": subsection["hidden"]
                                },
                                "parent_question": {
                                    "id": q_id,
                                    "text": question["text"],
                                    "description": question["description"],
                                    "format": question["format"],
                                    "mandatory": question["mandatory"],
                                    "hidden": question["hidden"]
                                },
                                "question": {
                                    "id": sq_id,
                                    "text": subquestion["text"],
                                    "description": subquestion["description"],
                                    "format": subquestion["format"],
                                    "mandatory": subquestion["mandatory"],
                                    "hidden": subquestion["hidden"]
                                }
                            }
        return answers

    def get_project_survey_data(self, id, expand=False, raise_for_status=True):
        """
        Get project survey data for the saved and draft profiles. Separate
        answer data into lists of common and unique answers.
        """

        # Get project.
        project = self.get_response_json(
            endpoint=self.endpoints["project"]
                         .replace("{project_id}", str(id)),
            parameters={},
            raise_for_status=raise_for_status
        )

        # Get saved project survey.
        survey = self.get_response_json(
            endpoint=self.endpoints["project_survey"]
                         .replace("{project_id}", str(id)),
            parameters={"expand": "answers"},
            raise_for_status=raise_for_status
        )

        # Get draft project survey. Include survey to ensure hidden answers are
        # returned (as `include=hidden` parameter does not impact the answers
        # list, just the full survey found in the `sections` field).
        survey_draft = self.get_response_json(
            endpoint=self.endpoints["project_survey_draft"]
                         .replace("{project_id}", str(id)),
            parameters={"include": "survey,hidden"},
            raise_for_status=raise_for_status
        )

        # Get answer dictionaries and ID lists.
        saved_answers = (
            sdeutils.get_dictionary_from_collection_with_ids(
                survey["answers"])
        )
        saved_answer_ids = list(saved_answers.keys())
        draft_answers = {}
        for section in survey_draft["sections"]:
            for subsection in section["subsections"]:
                for question in subsection["questions"]:
                    for answer in question["answers"]:
                        if answer["selected"]:
                            draft_answers[answer["id"]] = answer
                    for subquestion in question["subquestions"]:
                        for answer in subquestion["answers"]:
                            if answer["selected"]:
                                draft_answers[answer["id"]] = answer
        draft_answer_ids = list(draft_answers.keys())

        # Sort answer IDs into common and unique lists.
        common_answer_ids = list(
            set(saved_answer_ids).intersection(draft_answer_ids))
        unique_saved_answer_ids = list(
            set(saved_answer_ids) - set(draft_answer_ids))
        unique_draft_answer_ids = list(
            set(draft_answer_ids) - set(saved_answer_ids))

        # Expand answer data.
        if expand:

            library_answers = self.get_library_answers()
            def expand_answer(id):
                """
                Expand answer ID into dictionary.
                """
                answer = {
                    "id": id,
                    "text": library_answers[id]["text"],
                    "description": library_answers[id]["description"],
                    "hidden": library_answers[id]["hidden"],
                    "saved_selected_by": (
                        saved_answers[id]["selected_by"]
                        if id in saved_answers
                        else []
                    ),
                    "draft_selected_by": (
                        draft_answers[id]["selected_by"]
                        if id in draft_answers
                        else []
                    )
                }
                return answer

            # Populate dictionaries with expanded answer data.
            common_answers = {
                common_answer_id: expand_answer(common_answer_id)
                for common_answer_id in common_answer_ids
            }
            unique_saved_answers = {
                unique_saved_answer_id: expand_answer(unique_saved_answer_id)
                for unique_saved_answer_id in unique_saved_answer_ids
            }
            unique_draft_answers = {
                unique_draft_answer_id: expand_answer(unique_draft_answer_id)
                for unique_draft_answer_id in unique_draft_answer_ids
            }

        # Do not expand answers, just include ID lists.        
        else:
            common_answers = common_answer_ids
            unique_saved_answers = unique_saved_answer_ids
            unique_draft_answers = unique_draft_answer_ids

        # Assemble survey data dictionary and return.
        survey_data = {
            "common_answers": common_answers,
            "unique_saved_answers": unique_saved_answers,
            "unique_draft_answers": unique_draft_answers,
            "survey_complete": project["survey_complete"],
            "survey_dirty": project["survey_dirty"],
            "survey_locked": project["locked"],
            "survey_locked_by": project["locked_by"],
            "survey_locked_on": project["locked_on"],
            "profile_saved": project["profile"],
            "profile_draft": project["profile_draft"]
        }

        return survey_data

    def create_project_release(self, project_id, payload, raise_for_status=True):
        """
        Create a new project release from supplied project ID by POSTing
        supplied payload to project release endpoint.
        """
        release_project = self.post(
            self.endpoints["project_releases"]
                .replace("{project_id}", str(project_id)),
            {},
            payload,
            raise_for_status=raise_for_status
        )
        return release_project

    def _populate_global_role_cache(self):
        """
        Internal method to populate the global role cache.
        The group role field returned by the users endpoint includes the role
        name but not the ID so this cached data helps address the gap and
        prevent repeated API requests for the same global roles.
        """
        self.global_role_cache = (
            sdeutils.get_dictionary_from_collection_with_key(
                self.get_results(self.endpoints["global_roles"], {}),
                "name"
            )
        )

    def get_global_role(self, role_name, use_global_role_cache=True):
        """
        Get global role from supplied role name (case-sensitive). Cache
        global role data to speed up repeated lookups unless calling code
        overrides this default. Returns None if role name not found.
        """
        global_role = None

        # Use global role cache.
        if use_global_role_cache:

            # Get global roles and retain in cached dictionary. Obtain global
            # roles again if name not found in cache.
            if (not self.global_role_cache
                or role_name not in self.global_role_cache):
                    self._populate_global_role_cache()

            # Get role from cache. Default to None in case role_name is invalid.
            global_role = self.global_role_cache.get(role_name, None)

        # Do not use global role cache so make request and search for name.
        else:
            global_roles = self.get_results(self.endpoints["global_roles"], {})
            for global_role in global_roles:
                if global_role["name"] == role_name:
                    global_role = global_role["id"]
                    break

        return global_role

    def get_user_global_roles(self, user_id):
        """
        Get a list of a user's global roles from supplied user ID.
        """
        user_global_roles = []
        user = self.get_response_json(
            self.endpoints["user"].replace("{user_id}", str(user_id)), {})
        user_global_roles.append(self.get_global_role(user["role"]["name"]))
        for group in user["groups"]:
            group_role = self.get_global_role(group["role"])
            if group_role not in user_global_roles:
                user_global_roles.append(group_role)
            
        return user_global_roles

    def get_user_global_permissions(self, user_id):
        """
        Get list of a user's global permission flags.
        """
        user_global_permissions = []
        for role in self.get_user_global_roles(user_id):
            for permission in role["permissions"]:
                if permission not in user_global_permissions:
                    user_global_permissions.append(permission)
        return user_global_permissions

    def select_draft_answer(self, project_id, answer_id):
        """
        Select an answer ID on a project survey draft.
        Accepts project ID in string or integer format.
        Answer ID must be a string (e.g. "A1", "CA12").
        Answers that are not valid in the current draft state a 400 is returned
        and the patch method will raise an exception. Check for answer validity
        before trying to select it.
        """
        response = self.patch(
            self.endpoints["project_survey_draft_answer"]
                .replace("{project_id}", str(project_id))
                .replace("{answer_id}", answer_id),
            parameters={"include": "survey,hidden"},
            json_payload={"selected": True}
        )
        return response

    def clear_survey_draft(self, project_id):
        """
        Clear the project survey draft.
        """
        response = self.delete(
            self.endpoints["project_survey_draft"]
                .replace("{project_id}", str(project_id)),
            parameters={"include": "survey,hidden"}
        )
        return response

    def save_survey_draft(self, project_id):
        """
        Save the project survey draft.
        """
        response = self.post(
            self.endpoints["project_survey_draft"]
                .replace("{project_id}", str(project_id)),
            parameters={"include": "survey,hidden"},
            json_payload=None
        )
        return response

    def lock_project(self, project_id):
        """
        Patch project locked field with True and return success Boolean
        indicating whether the project with the specified ID is locked.
        Errors of any kind including invalid project ID, permission issues, or
        network failures will result in a False return value.
        """
        success = False
        try:
            response = self.patch(
                self.endpoints["project"]
                    .replace("{project_id}", str(project_id)),
                parameters={},
                json_payload={"project_locked": True}
            )
        except:
            pass
        else:
            if response.get("project_locked", False):
                success = True
        return success

    def unlock_project(self, project_id):
        """
        Patch project locked field with False and return success Boolean
        indicating whether the project with the specified ID is locked.
        Errors of any kind including invalid project ID, permission issues, or
        network failures will result in a False return value.
        """
        success = False
        try:
            response = self.patch(
                self.endpoints["project"]
                    .replace("{project_id}", str(project_id)),
                parameters={},
                json_payload={"project_locked": False}
            )
        except:
            pass
        else:
            if not response.get("project_locked", True):
                success = True
        return success

    def lock_project_survey(self, project_id):
        """
        Patch project survey locked field with True and return success Boolean
        indicating whether the survey of the project with the specified ID is
        locked.
        Errors of any kind including invalid project ID, permission issues, or
        network failures will result in a False return value.
        """
        # Note that the "locked" project key is deprecated and will be replaced
        # with "survey_locked" at some point in the future.
        success = False
        try:
            response = self.patch(
                self.endpoints["project"]
                    .replace("{project_id}", str(project_id)),
                parameters={},
                json_payload={"locked": True}
            )
        except:
            pass
        else:
            if response.get("locked", False):
                success = True
        return success

    def unlock_project_survey(self, project_id):
        """
        Patch project survey locked field with False and return success Boolean
        indicating whether the survey of the project with the specified ID is
        unlocked.
        Errors of any kind including invalid project ID, permission issues, or
        network failures will result in a False return value.
        """
        # Note that the "locked" project key is deprecated and will be replaced
        # with "survey_locked" at some point in the future.
        success = False
        try:
            response = self.patch(
                self.endpoints["project"]
                    .replace("{project_id}", str(project_id)),
                parameters={},
                json_payload={"locked": False}
            )
        except:
            pass
        else:
            if not response.get("locked", True):
                success = True
        return success

    def add_project_task_from_library(
        self, project_id, task_id, status_id=None, assigned_to=[], tags=[]):
        """
        Add a task to a project from the library.
        """
        payload = {
            "task_id": task_id,
            "assigned_to": assigned_to,
            "tags": tags
        }
        if status_id:
            payload["status_id"] = status_id
        response = self.post(
            endpoint=self.endpoints["project_tasks"]
                         .replace("{project_id}", str(project_id)),
            parameters={},
            json_payload=payload
        )
        return response

# Library import/export methods.

    def export_library(self, destination_folder, file_format):
        """
        Export libary zip file and save contents to destination folder.
        Method is synchronous so the code will wait until the file is downloaded
        or the timeout is exceeded.
        """
        success = False

        # Initiate export job and get job ID.
        job_payload = {"file_format": file_format}
        job = self.post(
            self.endpoints["library_export_jobs"],
            parameters={},
            json_payload=job_payload
        )
        job_id = job["id"]

        # Poll API to determine when the job is complete.
        job_done = False
        job_endpoint = (
            self.endpoints["library_export_job"]
                .replace("{library_export_job_id}", str(job_id))
        )
        for _ in range(self.library_export_retries + 1):
            time.sleep(self.library_export_delay)
            job_status = self.get_response_json(job_endpoint, parameters={})
            if job_status["status"] == "success":
                job_done = True
                break

        # Save the zip file.
        if job_done:
            artifacts_endpoint = (
                self.endpoints["library_export_job_artifacts"]
                    .replace("{library_export_job_id}", str(job_id))
            )
            file_path = (
                destination_folder + "sde_export.zip"
                if destination_folder.endswith("/")
                else destination_folder + "/" + "sde_export.zip"
            )
            with open(file_path, "wb") as file_writer:
                response = self.session.get(self.server_url + artifacts_endpoint)
                file_writer.write(response.content)
            success = True

        return success

# Cube API methods.

    def _cube_jwt_update_required(self):
        """
        Return Boolean indicating whether the Cube JWT has not been set or is
        too old and must be updated.
        """
        cube_jwt_update_required = False
        if (
            self.cube_jwt_updated is None
            or (
                (datetime.now() - self.cube_jwt_updated).total_seconds()
                > self.cube_jwt_max_age
            )
        ):
            cube_jwt_update_required = True
        return cube_jwt_update_required

    def _refresh_cube_jwt(self):
        """
        Obtains new Cube JWT and updates SdeClient properties if it requires
        an update.
        """
        if self._cube_jwt_update_required():
            self.cube_jwt = self.get_response_json(
                endpoint=self.cube_auth_endpoint,
                parameters={}
            )["token"]
            self.cube_jwt_updated = datetime.now()

    def get_cube_query_data(self, query):
        """
        Executes supplied Cube query and returns the data from the response.
        """
        self._refresh_cube_jwt()
        cube_headers = {
            "Authorization": f"{self.cube_jwt}",
            "Accept": "application/json"
        }
        query_response = requests.get(
            url=self.server_url + self.cube_query_endpoint,
            headers=cube_headers,
            params={"query": json.dumps(query)},
            verify=self.verify_certificate,
            proxies=self.proxies
        )
        return query_response.json()["data"]

    def get_cube_report_data(self, query_id):
        """
        Executes a query from an advanced report and returns the data from the
        response.
        """
        report = self.get_response_json(
            endpoint=self.endpoints["query"]
                         .replace("{query_id}", str(query_id)),
            parameters={}
        )
        query = report["query"]
        return self.get_cube_query_data(query)
