"""
SdeUtilities static class.
Updated: 2022-06-29
----
Change Log
2020-11-22 - Improved mixed_id_compare to support any number of alpha
             characters by using a regex.
2021-01-25 - Added get_dictionary_list_from_csv_file. Minor reformatting.
2021-03-25 - Fixed get_string_list_from_csv to return an empty list if the
             passed string is empty. Previous version returned a list containing
             a single empty string.
2021-10-07 - Specified UTF-8 file encoding for methods that write files.
2021-10-26 - Specified UTF-8-SIG file encoding for
             get_dictionary_list_from_csv_file method to ensure that CSV files
             with a single column don't cause the method to include the BOM
             u+FEFF whitespace character in the key name.
2021-12-14 - Added get_dictionary_from_json_file.
2022-03-07 - Added write_dictionary_list_to_csv_file.
2022-03-11 - Added extract_zip method.
             Added get_string_list_from_separated_string method.
             Modified get_string_list_from_csv to call the new generic method.
2022-04-27 - Added get_most_recent_notes method.
2022-06-29 - Modified write_json_to_file to expose the sort_keys parameter.
"""


import csv
import json
import os
import re
from zipfile import ZipFile


class sdeutils:
    """
    Collection of static utility methods for use with SdeClient.
    """

    @staticmethod
    def get_bool_from_string(boolean_string):
        """
        Return False unless parameter is case-insensitive variation of
        'true' or 't'.
        """
        boolean = False
        if boolean_string.lower() in ["true", "t"]:
            boolean = True
        return boolean

    @staticmethod
    def get_completion_percent(todo_count, done_count):
        """
        Get calculated completion percent.
        """
        completion_percent = 0
        todo_and_done_count = todo_count + done_count
        if todo_and_done_count > 0:
            completion_percent = round(100 
                                       * float(done_count)
                                       / float(todo_and_done_count))
        return completion_percent

    @staticmethod
    def write_report_lines_to_file(report_lines, output_filename):
        """
        Write list of string lists to CSV.
        """
        with open(output_filename, "w", encoding="utf-8", newline="") as output_file:
            output_file_writer = csv.writer(output_file)
            for report_line in report_lines:
                output_file_writer.writerow(report_line)

    @staticmethod
    def write_lines_to_file(lines, output_filename):
        """
        Write list of strings to a text file.
        """
        with open(output_filename, "w", encoding="utf-8", newline="") as output_file:
            for line in lines:
                output_file.write(line + "\n")

    @staticmethod
    def write_json_to_file(json_string, output_filename, sort_keys=True):
        """
        Write JSON in a readable format to a text file.
        """
        with open(output_filename, "w", encoding="utf-8") as output_file:
            json.dump(json_string, output_file, indent=2, sort_keys=sort_keys)

    @staticmethod
    def get_dictionary_from_json_file(input_filename):
        """
        Get contents of JSON file as dictionary.
        """
        with open(input_filename, encoding="utf-8") as input_file:
            dictionary = json.load(input_file)
        return dictionary

    @staticmethod
    def get_string_list_from_txt_file(input_filename):
        """
        Get list of strings from lines in a text file.
        """
        with open(input_filename) as input_file:
            file_lines = input_file.readlines()
        file_lines_in_list = [line.strip() for line in file_lines]
        return file_lines_in_list

    @staticmethod
    def get_dictionary_list_from_csv_file(input_filename):
        """
        Get dictionary from a CSV file. Strip whitespace in header rows
        to ensure key names don't contain whitespace.
        """
        with open(input_filename, encoding="utf-8-sig") as input_file:
            header = [field_name.strip()
                      for field_name
                      in input_file.readline().split(',')]
            dictionary_list = list(
                csv.DictReader(input_file, fieldnames=header)
            )
        return dictionary_list

    @staticmethod
    def write_dictionary_list_to_csv_file(dictionary_list, output_filename, keys=None):
        """
        Write dictionary list to a CSV file.
        """
        if not keys:
            keys = dictionary_list[0].keys()
        with open(output_filename, "w", encoding="utf-8", newline="") as output_file:
            output_file_writer = csv.DictWriter(output_file, keys)
            output_file_writer.writeheader()
            output_file_writer.writerows(dictionary_list)
  
    @staticmethod
    def get_string_list_from_separated_string(separated_string, separator):
        """
        Get list of string values from a string separated by one or more characters.
        Returns empty list when separated_string is empty.
        """
        string_as_list = []
        if separator in str(separated_string):
            for value in separated_string.split(separator):
                string_as_list.append(value.strip())
        elif separated_string:
            string_as_list.append(separated_string)
        return string_as_list

    @staticmethod
    def get_string_list_from_csv(csv_string):
        """
        Get list of string values from CSV string.
        Returns empty list when csv_string is empty.
        """
        return sdeutils.get_string_list_from_separated_string(csv_string, ",")

    @staticmethod
    def get_int_list_from_csv(csv_string):
        """
        Get list of int values from CSV string.
        """
        return [int(value)
                for value
                in sdeutils.get_string_list_from_csv(csv_string)]

    @staticmethod
    def get_unique_item_list_from_lists(first_list, second_list):
        """
        Get a list of unique items from two lists that may included duplicates.
        """
        first_set = set(first_list)
        second_set = set(second_list)
        return list(first_set) + list(second_set - first_set)

    @staticmethod
    def get_dictionary_from_collection_with_ids(collection):
        """
        Get dictionary of objects with "id" value as key.
        """
        return sdeutils.get_dictionary_from_collection_with_key(
            collection, "id"
        )

    @staticmethod
    def get_dictionary_from_collection_with_key(collection, key_name):
        """
        Get dictionary of objects with key_name value as key.
        """
        collection_dictionary = {}
        for item in collection:
            collection_dictionary[item[key_name]] = item
        return collection_dictionary

    @staticmethod
    def mixed_id_compare(id1, id2):
        """
        Comparison function used for sorting by IDs of the same object type
        that mix letters and ints. Put custom IDs starting with "C" first.
        """
        id1_is_custom = (id1[:1] == "C")
        id2_is_custom = (id2[:1] == "C")
        if id1_is_custom and not id2_is_custom:
            return -1
        elif not id1_is_custom and id2_is_custom:
            return 1
        id1_int_portion = int(re.findall(r"\d+|\D+", id1)[1])
        id2_int_portion = int(re.findall(r"\d+|\D+", id2)[1])
        if id1_int_portion < id2_int_portion:
            return -1
        elif id1_int_portion == id2_int_portion:
            return 0
        else:
            return 1

    @staticmethod
    def extract_zip(file_path, extract_path=None):
        """
        Extract contents of a zip file. If extract_path is omitted, extract
        contents to the folder where the file exists.
        """
        if not extract_path:
            extract_path = os.path.dirname(file_path)
        with ZipFile(file_path, "r") as zip_file:
            zip_file.extractall(path=extract_path)

    @staticmethod
    def get_most_recent_notes(note_collection):
        """
        Get dictionary of task or verification notes keyed by user email. Supply
        a collection of notes returned by the API and they are searched to find
        the most recent note for each unique email.
        """
        recent_notes = {}
        for note in note_collection:
            updater_email = note["updater"]["email"]
            if (updater_email not in recent_notes
                or note["created"] > recent_notes[updater_email]["created"]):
                    recent_notes[updater_email] = note
        return recent_notes
