"""
Log script progress messages to the console and/or text file.
Created: 2020-05-13
Last Updated: 2023-03-09
Author: Scott McDowell (smcdowell@securitycompass.com)

Change Log
2020-05-13 - Initial version.
2020-05-15 - Fix for output_filename assertion in initializer method.
2020-06-05 - Added option to print timestamps.
2021-05-28 - Improve header text block, comments. No functional changes.
2022-01-19 - Fixed default output_filename formatting.
2022-02-08 - Added debug, notify, warn, and exception methods to improve the
             readability of the code that uses this class. They use the original
             add_line method which is retained for backward compatibilty.
2023-03-09 - Check local environment language encoding and ensure that the line
             that's printed to the console matches the user's environment.
             This better supports environments where LANG=en_US.UTF-8 is not
             used, for example LANG=en_US (ISO8859-1/latin-1).
"""


import sys
from datetime import datetime

from sdeutilities import sdeutils


class simpleLog:
    """
    Simple text log class.
    """

    def __init__(self, debug_mode=False, output_filename=None,
                 print_timestamps=True):
        assert isinstance(debug_mode, bool)
        assert output_filename is None or isinstance(output_filename, str)
        assert isinstance(print_timestamps, bool)

        self.debug_mode = debug_mode
        self.output_filename = (
            output_filename if output_filename is not None
            else datetime.now().strftime("%Y-%m-%d_%H-%M-%S") + "_log.txt"
        )
        self.lines = []
        self.print_timestamps = print_timestamps

    def add_line(self, log_level, line):
        """
        Add a line to the log.
        """
        assert isinstance(log_level, int)
        assert isinstance(line, str)

        if log_level == 1 or self.debug_mode:
            if self.print_timestamps:
                log_line = (datetime.now().strftime("%Y-%m-%d %H:%M:%S ")
                            + line)
            else:
                log_line = line

            # Ensure the string to print is encoded for the local environment.
            if sys.stdout.encoding.lower() != "utf-8":
                log_line = (
                    log_line.encode("utf-8")
                            .decode(sys.stdout.encoding, errors="replace")
                )
            print(log_line)

        self.lines.append([
            str(log_level),
            datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            line
        ])
    
    def debug(self, line):
        """
        Add a line with log level 2.
        """
        self.add_line(2, line)

    def notify(self, line):
        """
        Add a line with log level 1.
        """
        self.add_line(1, line)

    def warn(self, line):
        """
        Add a warning line (prefix "Warning: ") with log level 1.
        """
        self.add_line(1, f"Warning: {line}")

    def exception(self, exception):
        """
        Add an exception line (prefix "Exception: ") with log level 1.
        """
        self.add_line(1, f"Exception: {exception}")

    def write_log_file(self):
        """
        Write contents of the log to a CSV file.
        """
        if self.output_filename is not None:
            sdeutils.write_report_lines_to_file(self.lines,
                                                    self.output_filename)
